
import java.awt.Point;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import java.util.HashMap;
import java.util.Map;

import javax.swing.JDialog;
import javax.swing.JOptionPane;


/*******************************************************************************
 * Knihovní třída {@code IO} obsahuje sadu metod
 * pro jednoduchý vstup a výstup prostřednictvím dialogovýách oken
 * spolu s metodou zastavující běh programu na daný počet milisekund
 * a metodu převádějící texty na ASCII jednoduchým odstraněním diakritiky.
 *
 * @author   Rudolf PECINOVSKÝ
 * @version  3.00.001
 */
public final class IO
{
//== KONSTANTNÍ ATRIBUTY TŘÍDY =================================================
//== PROMĚNNÉ ATRIBUTY TŘÍDY ===================================================

    /** Pozice dialogových oken. */
    private static Point poziceOken = new Point(0,0);


//== STATICKÝ INICIALIZAČNÍ BLOK - STATICKÝ KONSTRUKTOR ========================
//== KONSTANTNÍ ATRIBUTY INSTANCÍ ==============================================
//== PROMĚNNÉ ATRIBUTY INSTANCÍ ================================================
//== PŘÍSTUPOVÉ METODY VLASTNOSTÍ TŘÍDY ========================================
//== OSTATNÍ NESOUKROMÉ METODY TŘÍDY ===========================================

    /***************************************************************************
     * Počká zadaný počet milisekund.
     * Na přerušení nijak zvlášť nereaguje - pouze skončí dřív.
     * Před tím však nastaví příznak, aby volající metoda poznala,
     * že vlákno bylo žádáno o přerušení.
     *
     * @param milisekund   Počet milisekund, po něž se má čekat.
     */
    public static void čekej( int milisekund )
    {
        try {
            Thread.sleep( milisekund);
        }catch( InterruptedException e) {
            Thread.currentThread().interrupt();
        }
    }


    /***************************************************************************
     * Při splnění zadané podmínky otevře dialogové okno s nápisem KONEC
     * a po jeho zavření ukončí program.
     *
     * @param platí  Podmínka, po jejímž splnění se program ukončí
     */
    public static void konecKdyž(boolean platí)
    {
        konecKdyž(platí, null);
    }


    /***************************************************************************
     * Při splnění zadané podmínky otevře dialogové okno se zadanou zprávou
     * a po jeho zavření ukončí program.
     *
     * @param platí  Podmínka, po jejímž splnění se program ukončí
     * @param zpráva Zpráva vypisovaná v dialogovém okně. Je-li {@code null}
     *               nebo prázdný řetězec, vypíše <b>{@code KONEC}</b>.
     */
    public static void konecKdyž(boolean platí, String zpráva)
    {
        if (platí) {
            if ((zpráva == null)  ||  (zpráva.equals(""))) {
                zpráva = "KONEC";
            }
            zpráva(zpráva);
            System.exit(0);
        }
    }


    /***************************************************************************
     * Zbaví zadaný text diakritických znamének; současně ale odstraní také
     * všechny další znaky nespadající do tabulky ASCII.
     *
     * @param text Text určený k "odháčkování"
     * @return  "Odháčkovaný" text
     */
    public static String odháčkuj( String text )
    {
        return Odháčkuj.text(text);
    }


    /***************************************************************************
     * Nastaví pozici příštího dialogového okna.
     *
     * @param x  Vodorovná souřadnice
     * @param y  Svislá souřadnice
     */
    public static void oknaNa( int x, int y )
    {
        poziceOken = new Point( x, y );
    }


    /***************************************************************************
     * Zobrazí dialogové okno se zprávou a umožní uživateli odpovědět
     * ANO nebo NE. Vrátí informaci o tom, jak uživatel odpověděl.
     * Neodpoví-li a zavře dialog, ukončí program.
     *
     * @param dotaz   Zobrazovaný text otázky.
     * @return <b>{@code true}</b> Odpověděl-li uživatel <b>ANO</b>,
     *         <b>{@code false}</b> odpověděl-li <b>NE</b>
     */
    public static boolean souhlas( Object dotaz )
    {
        JOptionPane jop = new JOptionPane(
                                dotaz,
                                JOptionPane.QUESTION_MESSAGE,   //Message type
                                JOptionPane.YES_NO_OPTION       //Option type
                                );
        processJOP( jop );
        int answer = (Integer)jop.getValue();
        return (answer == JOptionPane.YES_OPTION);
    }


    /***************************************************************************
     * Zobrazí dialogové okno s výzvou k zadání reálné hodoty;
     * při zavření okna zavíracím tlačítkem ukončí aplikaci.
     *
     * @param výzva        Text, který se uživateli zobrazí.
     * @param doubleImpl   Implicitní hodnota.
     * @return Uživatelem zadaná hodnota, resp. potvrzená implicitní hodnota.
     */
    public static double zadej( Object výzva, double doubleImpl )
    {
        return Double.parseDouble( zadej( výzva, ""+doubleImpl ).trim() );
    }


    /***************************************************************************
     * Zobrazí dialogové okno s výzvou k zadání celočíselné hodoty;
     * při zavření okna nebo stisku tlačítka Cancel ukončí aplikaci.
     *
     * @param výzva     Text, který se uživateli zobrazí.
     * @param intImpl   Implicitní hodnota.
     * @return Uživatelem zadaná hodnota, resp. potvrzená implicitní hodnota.
     */
    public static int zadej( Object výzva, int intImpl )
    {
        return Integer.parseInt( zadej( výzva, ""+intImpl ).trim() );
    }


    /***************************************************************************
     * Zobrazí dialogové okno s výzvou k zadání textové hodoty;
     * při zavření okna nebo stisku tlačítka Cancel ukončí aplikaci.
     *
     * @param výzva        Text, který se uživateli zobrazí.
     * @param stringImpl   Implicitní hodnota.
     * @return Uživatelem zadaná hodnota, resp. potvrzená implicitní hodnota.
     */
    public static String zadej( Object výzva, String stringImpl )
    {
        JOptionPane jop = new JOptionPane(
                              výzva,
                              JOptionPane.QUESTION_MESSAGE,   //Message type
                              JOptionPane.DEFAULT_OPTION  //Option type - OK
                              );
        jop.setWantsInput(true);
        jop.setInitialSelectionValue(stringImpl);
        processJOP(jop);
        String answer = jop.getInputValue().toString();
        return answer;
    }


    /***************************************************************************
     * Zobrazí dialogové okno se zprávou a počká, až je uživatel odklepne;
     * při zavření okna ukončí aplikaci.
     *
     * @param text   Zobrazovaný text.
     */
    public static void zpráva( Object text )
    {
        JOptionPane jop = new JOptionPane(
                          text,                            //Sended message
                          JOptionPane.INFORMATION_MESSAGE  //Message type
                          );
        processJOP( jop );
    }



//##############################################################################
//== KONSTRUKTORY A TOVÁRNÍ METODY =============================================

    /***************************************************************************
     * Třída IO je knihovní třídou a proto není určena k tomu,
     * aby měla nějaké instance.
     */
    private IO() {}


//== ABSTRAKTNÍ METODY =========================================================
//== PŘÍSTUPOVÉ METODY VLASTNOSTÍ INSTANCÍ =====================================
//== OSTATNÍ NESOUKROMÉ METODY INSTANCÍ ========================================
//== SOUKROMÉ A POMOCNÉ METODY TŘÍDY ===========================================

    /***************************************************************************
     * Creates a dialog from the given {@link JOptionPane}, makes it non-modal
     * and waits for its closing leaving the entered value in the parameter's
     * attribute {@code value}. If the user closed the dialog
     * from the window's system menu, exit the whole application.
     */
    private static void processJOP( JOptionPane jop )
    {
        final int WAITING=0, CANCELLED=1;
        final Boolean[] USER = {true, false};

        final JDialog jd = jop.createDialog((JDialog)null, "Information"  );

        jd.addWindowListener( new WindowAdapter()
        {
            /** Set the information about closing the window from its
             *  systme menu - the application will be cancelled. */
            @Override
            public void windowClosing(WindowEvent e) {
                synchronized( USER ) {
                    USER[CANCELLED] = true;
                    System.exit( 1 );
                }
            }
            @Override
            public void windowDeactivated(WindowEvent e) {
                poziceOken = jd.getLocation();
                if( jd.isShowing() ) {
                    return;
                }else{
                    jd.dispose();
                    synchronized( USER ) {
                        USER[WAITING] = false;
                        USER.notifyAll();
                    }
                }
            }
         });

        jd.setModal( false );
        jd.setVisible( true );
        jd.setLocation( poziceOken  );
        jd.toFront();
        jd.setAlwaysOnTop(true);
//        jd.setAlwaysOnTop(false);

        //Waiting until the user answers or closes the dialog
        synchronized( USER ) {
            while( USER[WAITING] ) {
                try {
                    USER.wait();
                } catch (InterruptedException ie ) {
                    Thread.currentThread().interrupt();
                }
            }
        }
    }



//== SOUKROMÉ A POMOCNÉ METODY INSTANCÍ ========================================
//== INTERNÍ DATOVÉ TYPY =======================================================

    /***************************************************************************
     * Třída {@code Odháčkuj_RUP} je knihovní třídou poskytující metodu na
     * odstranění diakritiky ze zadaného textu a následné převedení všech znaků,
     * jejichž kód je stále větší než 127, na příslušné kódové
     * únikové posloupnosti (escape sekvence).
     */
    private static class Odháčkuj
    {
    //== KONSTANTNÍ ATRIBUTY TŘÍDY =============================================

        /** Mapa s převody znaků do ASCII. */
        private static final Map<Character,String> PŘEVOD =
                                            new HashMap<Character, String>(64);
    //== PROMĚNNÉ ATRIBUTY TŘÍDY ===============================================
    //== STATICKÝ INICIALIZAČNÍ BLOK - STATICKÝ KONSTRUKTOR ====================
        static {
            String[][] dvojice = {

                {"Á", "A"},  {"á", "a"},    {"Ä", "AE"}, {"ä", "ae"},
                {"Č", "C"},  {"č", "c"},
                {"Ď", "D"},  {"ď", "d"},    {"Ë", "E"},  {"ë", "e"},
                {"É", "E"},  {"é", "e"},
                {"Ě", "E"},  {"ě", "e"},
                {"Í", "I"},  {"í", "i"},    {"Ï", "IE"}, {"ï", "ie"},
                {"Ĺ", "L"},  {"ĺ", "l"},    {"Ľ", "L"},  {"ľ", "l"},
                {"Ň", "N"},  {"ň", "n"},
                {"Ó", "O"},  {"ó", "o"},    {"Ö", "OE"}, {"ö", "oe"},
                {"Ô", "O"},  {"ô", "o"},
                {"Ŕ", "R"},  {"ŕ", "r"},    {"Ř", "R"},  {"ř", "r"},
                {"Š", "S"},  {"š", "s"},
                {"Ť", "T"},  {"ť", "t"},
                {"Ú", "U"},  {"ú", "u"},    {"Ü", "UE"}, {"ü", "ue"},
                {"Ů", "U"},  {"ů", "u"},
                {"Ý", "Y"},  {"ý", "y"},    {"Ÿ", "YE"}, {"ÿ", "ye"},
                {"Ž", "Z"},  {"ž", "z"},
                {"ß", "ss"},
                {"«", "<<"}, {"»", ">>"},
//                {"",""},
            };
            for( String[] ss : dvojice ) {
                PŘEVOD.put( new Character(ss[0].charAt(0)),  ss[1] );
            }
            dvojice = null;
        }



    //== KONSTANTNÍ ATRIBUTY INSTANCÍ ==========================================
    //== PROMĚNNÉ ATRIBUTY INSTANCÍ ============================================
    //== PŘÍSTUPOVÉ METODY VLASTNOSTÍ TŘÍDY ====================================
    //== OSTATNÍ NESOUKROMÉ METODY TŘÍDY =======================================

        /***********************************************************************
         * Zbaví zadaný text diakritických znamének - <b>POZOR</b> -
         * Spolu s nimi odstraní také všechny znaky s kódem větším než 127.
         *
         * @param text Text určený k "odháčkování"
         * @return  "Odháčkovaný" text
         */
        public static String text( CharSequence text )
        {
            final int DÉLKA = text.length();
            final StringBuilder sb = new StringBuilder(DÉLKA);
            for( int i = 0;   i < DÉLKA;   i++ ) {
                char c = text.charAt(i);
                if( c < 128 ) {
                    sb.append(c);
                }else if( PŘEVOD.containsKey(c) ) {
                    sb.append( PŘEVOD.get(c) );
                }else {
                    sb.append( rozepiš(c) );
                }
            }
            return sb.toString();
        }



    //##########################################################################
    //== KONSTRUKTORY A TOVÁRNÍ METODY =========================================

       /** Soukromy konstruktor bránící vytvoření instance. */
        private Odháčkuj() {}


    //== ABSTRAKTNÍ METODY =====================================================
    //== PŘÍSTUPOVÉ METODY VLASTNOSTÍ INSTANCÍ =================================
    //== OSTATNÍ NESOUKROMÉ METODY INSTANCÍ ====================================
    //== SOUKROMÉ A POMOCNÉ METODY TŘÍDY =======================================

        /***********************************************************************
         * Rozepíše zadaný znak do příslušné ńikové k´dové posloupnosti.
         *
         * @param c Převáděný znak
         * @return Text ve formátu \\uXXXX
         */
        private static String rozepiš(char c) {
            return String.format( "\\u%04x", (int)c );
        }



    //== SOUKROMÉ A POMOCNÉ METODY INSTANCÍ ====================================
    //== INTERNÍ DATOVÉ TYPY ===================================================
    //== TESTY A METODA MAIN ===================================================
    }



//== TESTY A METODA MAIN =======================================================
}
